/*
A GLUT program using OpenGL, GLUI, and GLSL.

Orion Sky Lawlor, olawlor@acm.org, 2007/01/18 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=5.0; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 model1Mat(1.0), model2Mat(1.0), model3Mat(1.0), viewMat(1.0);

/* Draw "the world": everything in our scene */
void draw_world(void) {
	glPushMatrix();
	/* Segment 1 */
	glMultMatrixf(&model1Mat[0][0]);
	glTranslatef(0,0.5,0);
	glColor4f(1,1,0,1); /* yellow */
	glutSolidTorus(0.2,0.5,40,60);
	glTranslatef(0,0.5,0);
	
	/* Segment 2 */
	glMultMatrixf(&model2Mat[0][0]);
	glTranslatef(0,0.5,0);
	glColor4f(1,0,0,1); /* red */
	glutSolidTorus(0.2,0.5,40,60);
	glTranslatef(0,0.5,0);
	
	/* Segment 3 */
	glMultMatrixf(&model3Mat[0][0]);
	glTranslatef(0,0.5,0);
	glColor4f(1,0,1,1); /* purple */
	glutSolidTorus(0.2,0.5,40,60);
	glTranslatef(0,0.5,0);
	
	/* End--a teapot */
	glTranslatef(0,0.5,0);
	glColor4f(0,0,1,1); /* blue */
	glutSolidTeapot(0.5);
	
	glPopMatrix();
}

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	// Clear the color and depth buffers
	glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	/* Depth testing */
	glEnable(GL_DEPTH_TEST);
	
	/* Alpha blending */
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	/* if you're not Dr. Lawlor, you'd switch to GL_MODELVIEW here */
	
	mat4 m;
	glGetFloatv(GL_PROJECTION_MATRIX,&m[0][0]); /* read back the projection matrix */
	mat4 t=mat4(1.0); /* translation matrix */
	t[3].z=-distance; /* set translation offset */
	m=m*t; /* multiply old matrix by translation matrix */
	glLoadMatrixf(&m[0][0]);
	
	//glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: can also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Set up programmable shader. */
	static oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	p->begin();
	p->set("colorscale",1.0f); /* normal case-- don't mess with colors */
	
	/* Draw a big green ground polygon */
	double ground=100.0; /* extent of ground in X and Z */
	double groundY=0.0;
	glColor4f(0.1,0.4,0.1,1); /* green */
	glNormal3f(0,1,0); /* facing straight up */
	oglTextureQuad(
		oglVector3d(-ground,groundY,-ground), oglVector3d(+ground,groundY,-ground), 
		oglVector3d(-ground,groundY,+ground), oglVector3d(+ground,groundY,+ground)
	);
	
	// Draw the world 
	draw_world();
	
	// Flatten the projection matrix along Y, and draw the world again in black
	glTranslatef(0,groundY+0.005,0);
	glScalef(1.0,0.0,1.0);
	p->set("colorscale",0.0f); /* normal case-- scale all colors by 0.0 */
	draw_world();
	
	p->end();
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

void main_motion(int x,int y) {
	printf("Mouse is at %d, %d\n",x,y);
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	glutMotionFunc  (main_motion);
	
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);
	/* This causes the main window to be redrawn when a GLUI control changes */
	glui->set_main_gfx_window( main_window );

	GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "Distance:", &distance);
	spin_zoom->set_float_limits(0.0, 10.0);
	
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Joint1:",&model1Mat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Joint2:",&model2Mat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Joint3:",&model3Mat[0][0]);
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
